#!/usr/bin/env python
# Jack Gallimore 2016
import numpy as np
import matplotlib.pyplot as mp
# DREAMZS is the MCMC stepper. We need it.
from dreamZPT.wrapper import DREAMZS
from dreamZPT.fastKDE import fast_kde

'''Documentation(-ish) for DREAMZS

Based on the DREAM(ZS) MCMC algorithm of ter Braak and Vrugt,
"Differential Evolution Markov Chain with snooker updater and fewer
chains," Statistics and Computing, December 2008, Volume 18, Issue 4,
pp 435-446.

I've written a python version of the DREAM(ZS) algorithm for MCMC
model fitting analysis. To use it, first make sure that the dreamZPT
folder is in your working area. To perform the MCMC analysis, you need
to import the DREAMZS function:

from dreamZPT.wrapper import DREAMZS

You'll need to define at least two support functions for your model.

(1) A function that calculates the log probability (= -chi-squared/2.0
) for your fit. The arguments should be the parameter array for the
model (p), and any information about the data, for example (x, y, u) =
independent measurements, dependent measurements (that you are trying
to model), and the uncertainties of the dependent measurements.

Example:

def logProbability(p, x, y, u):
    m = # expression that calculates the model from p and x
    normResid = (m-y)/u
    chisq = sum(normResid**2)
    return -chisq/2.0

(2) A function that calculates the log probability using prior
information about your parameters. For example, if you know a
parameter is > 0.0, you could have a statement like the following,

def logPrior(p): # p is a parameter array
    if p[0] < 0.:
        return -inf
    return 0.0

If you don't have prior information about your parameters, just define
a stub function that returns 0.

Next, you have to define an array of initial guesses. By convention, I
usually call it the Z array. The following snippet of code sets up
parameters randomly distributed between 0 and 1, but you can be more
sophisticated in setting up these parameters.

# In this example, create a Z array of 20 initial guesses for 3
# parameters, with values between 0 and 1.

from numpy.random import uniform
Z = uniform(size=(20,4))

Now you can call the DREAMZS function.

Zfinal, ZP, successVec = DREAMZS(Z, logProbability, logPrior, 10000, args=(x,y,u))

Inputs:
=======
Z = initial guesses

logProbability is the function that evaluates the probability of the
model being a good fit

logPrior calculates probabilities based on prior information about the
parameters.

The number of iterations (here, 10000). Note that the number of
function evaluations using default parameters is 30 * number of
iterations.

args=(...) additional arguments to the logProbability function beyond
the parameter array, which must be the first argument.

Outputs
=======
Zfinal = complete history of MCMC samples
Example: Zfinal[:,0] = all of the samples for the first parameters
Z[400] = Accepted parameters p[0], p[1], ... for the 400th iteration.

ZP = the log probability (to within an additive constant) for that set
of model parameters. For example, the chisq value of the 512th
iteration = -ZP[512] / 2.0.

successVec = An array of zeros and ones, summarizing whether an MCMC
step was successful or not. Used for debugging and diagnostics.

Feeling the Burn
================
There's a convergence time, or number of iterations, associated with
any non-linear fitting technique. For MCMC, it's called the 'burn-in'
period. Basically, the first few (or many) iterations are out of
equilibrium, and they need to be discarded before analyzing the
results. Visual inspection is usually good enough to identify the
burn-in period, but the dreamZPT library provides an automated
technique.

from dreamZPT.diagnostics import autoburn
burn = autoburn(ZP)

The function autoburn examines the ZP array (equivalently, the
chi-squared array) to identify where the chi-squared values settle to
an equilibrium. It then returns (an estimate of) the first array
element where the fitting seems to have converged.

Clever Tricks
=============
The mean and uncertainty of a parameter can be calculated
trivially. For example, for parameter #2:

p2 = mean(Z[burn:,2]) # mean
up2 = std(Z[burn:,2]) # uncertainty

I'll show some other tricks in class.

Advantages and Disadvantages of MCMC
==================================== 
Disadvantage: It's slow; however, computers are getting embarrasingly
fast, and MCMC is state-of-the-art these days.

Disadvantage: For every problem, there is an optimal stepping
technique, and you won't know what it is a priori. Nonetheless,
DREAM(ZS) is pretty robust to anything you throw at it.

Advantage: Unlike other non-linear approaches, you get a best-fit
value for parameters *and* their distributions.

'''

# need to define a model function.
# Arguments are (parameters, independent variable values)
def model(p, x):
    Amplitude = p[0]
    Period = p[1]
    Phase = p[2]
    Decay = p[3]
    y = Amplitude * np.sin(2*np.pi*x/Period + Phase)*np.exp(-x/Decay)
    return y

# logProbability calculates chi-squared, returns log probability
def logProbability(p, x, y, u):
    m = model(p, x)
    chisq = sum((y-m)**2/u**2)
    return -chisq/2.0

# Use prior info to constrain parameters. If no prior info, this function should return 0.
def logPrior(p): 
    Amplitude = p[0]
    Period = p[1]
    Phase = p[2]
    Decay = p[3]
    if Amplitude > 2.0: # from inspection the amplitude is < 2
        return -np.inf
    if Amplitude < 0.0:
        return -np.inf
    if Decay <= 0.0: # Avoid divide by zero issues
        return -np.inf
    if Phase > 2.0*np.pi: # keep the phases between 0 and 2 pi
        return -np.inf
    if Phase < 0.0:
        return -np.inf
    if Period <= 0.0:
        return -np.inf
    return 0.0

# Read in data
x, y, u = np.loadtxt('dsc.txt', unpack=True)

'''
Initial guesses
Each row of Z is a set of initial guesses for the parameters
'''
# low values for (Amplitude, Period, Phase, Decay)
plow = np.array([0.0, 0.0, 0.0, 0.0001]) 
# high values for (Amplitude, Period, Phase, Decay)
phigh = np.array([1.0, 3.0, np.pi*2.0, 5.0]) 
# initially, 20 guesses for four parameters
Z = np.zeros((20,4)) 


#Generate 20 initial guesses, uniformly distributed between plow and phigh

for i in range(20):
    Z[i] = plow + (phigh - plow)*np.random.uniform(size=4) 

'''
Now run MCMC
Arguments are:
* Z (initial guesses)
* logProbability (name of the function that calculates -chisq/2.0)
* logPrior (name of function that calculates parameter constraints)
Number of iterations (10000, here)
The data (passed in the variable 'args')
'''
Z, ZP, successVec = DREAMZS(Z, logProbability, logPrior, 10000, args=(x,y,u))

# do some analysis
from dreamZPT.diagnostics import autoburn

burn = autoburn(ZP) # autoburn estimates the iteration where solutions have converged

# plot an example model fit
p = Z[-1] # pick the last parameter set; could also use the one with the lowest ZP
m = model(p, x)
mp.figure(0)
mp.clf()
mp.errorbar(x, y, u, fmt='k.')
mp.plot(x, m, 'r-')
mp.xlabel('Time')
mp.ylabel('Signal')
mp.savefig('modelFit.pdf')

# plot the amplitude trace
Amplitude = Z[burn:,0]
mp.figure(1)
mp.clf()
mp.plot(Amplitude, '.')
mp.xlabel('Iteration')
mp.ylabel('Amplitude')
mp.draw()
mp.savefig('amplitudeTrace.pdf')

# make a histogram
mp.figure(2)
mp.clf()
mp.hist(Amplitude, bins=30)
mp.xlabel('Amplitude')
mp.ylabel('Count')
mp.draw()
mp.savefig('amplitudeHist.pdf')

# Plot two variables against one another
Period = Z[burn:,1]
Phase = Z[burn:,2]
Decay = Z[burn:,3]

# First, amplitude and decay
mp.figure(3)
mp.clf()
mp.plot(Amplitude, Decay, ',')
mp.xlabel('Amplitude')
mp.ylabel('Decay')
mp.draw()
mp.savefig('amplitudeDecay.pdf')

# Next, amplitude and phase
mp.figure(3)
mp.clf()
mp.plot(Amplitude, Phase, ',')
mp.xlabel('Amplitude')
mp.ylabel('Decay')
mp.draw()
mp.savefig('amplitudePhase.pdf')

'''
Notice that the amplitude and decay are covariant!
Amplitude and phase are not covariant.
'''


'''
Generate a plot of Chi-squared
'''
chisq = -2.0*ZP[burn:]
mp.figure(4)
mp.hist(chisq, bins=30)
mp.ylabel('Count')
mp.xlabel('$\\chi^2$')
mp.draw()
mp.savefig('chisq.pdf')


# Calculate some basic statistics
print 'Amplitude = ', np.mean(Amplitude), ' +/- ', np.std(Amplitude)
print 'Period = ', np.mean(Period), ' +/- ', np.std(Period)
print 'Phase = ', np.mean(Phase), ' +/- ', np.std(Phase)
print 'Decay = ', np.mean(Decay), ' +/- ', np.std(Decay)

'''
Generate Probability contours.
Calculate curves that enclose 68% of the data, 95%, and 99.7%, 
corresponding to 1, 2, 3 sigma. These are equivalent to changing
chi-squared by +/- 1, 4, & 9.
'''
# use a kernel density estimator to calculate the density of samples
density, extent = fast_kde(Amplitude, Decay, gridsize=(200,200))
# calculate the cumulative distribution function
dr = density.ravel()
H = np.sort(dr)[::-1]
cdf = np.array([np.sum(H[H > x])/H.sum() for x in H])
# convert to percentiles
from scipy.interpolate import interp1d
fcdf = interp1d(H, cdf, bounds_error=False)
percentileMap = np.reshape(fcdf(dr), density.shape) * 100.

# Contour levels for 1, 2, 3 sigma
levels = [68.27, 95.45, 99.73]
mp.figure(5)
mp.clf()
cs = mp.contour(percentileMap, levels=levels, extent=extent, colors='k')
mp.clabel(cs, inline=1, fontsize=10)
mp.xlabel('Amplitude')
mp.ylabel('Decay Time')
mp.savefig('contourAmplitudeDecay.pdf')


